# Generated by Django 4.1 on 2022-08-21 20:06

from django.db import migrations, models

import signal_webhooks.fields
import signal_webhooks.utils


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Webhook",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        db_index=True,
                        help_text="Webhook name.",
                        max_length=256,
                        unique=True,
                        verbose_name="name",
                    ),
                ),
                (
                    "signal",
                    models.IntegerField(
                        choices=[
                            (0, "CREATE"),
                            (1, "UPDATE"),
                            (2, "DELETE"),
                            (3, "CREATE OR UPDATE"),
                            (4, "CREATE OR DELETE"),
                            (5, "UPDATE OR DELETE"),
                            (6, "ALL"),
                        ],
                        help_text="Signal the webhook fires to.",
                        verbose_name="signal",
                    ),
                ),
                (
                    "ref",
                    models.CharField(
                        db_index=True,
                        help_text="Dot import notation to the model the webhook is for.",
                        max_length=1024,
                        validators=[signal_webhooks.utils.model_from_reference],
                        verbose_name="referenced model",
                    ),
                ),
                (
                    "endpoint",
                    models.URLField(
                        help_text="Target endpoint for this webhook.",
                        max_length=2048,
                        verbose_name="endpoint",
                    ),
                ),
                (
                    "headers",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        help_text="Headers to send with the webhook request.",
                        validators=[signal_webhooks.utils.is_dict],
                        verbose_name="headers",
                    ),
                ),
                (
                    "auth_token",
                    signal_webhooks.fields.TokenField(
                        blank=True,
                        default="",
                        help_text="Authentication token to use in an Authorization header.",
                        max_length=8000,
                        validators=[signal_webhooks.utils.decode_cipher_key],
                        verbose_name="authentication token",
                    ),
                ),
                (
                    "enabled",
                    models.BooleanField(
                        default=True,
                        help_text="Is this webhook enabled?",
                        verbose_name="enabled",
                    ),
                ),
                (
                    "keep_last_response",
                    models.BooleanField(
                        default=False,
                        help_text="Should the webhook keep a log of the latest response it got?",
                        verbose_name="keep last response",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="When the webhook was created.",
                        verbose_name="created",
                    ),
                ),
                (
                    "updated",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="When the webhook was last updated.",
                        verbose_name="updated",
                    ),
                ),
                (
                    "last_response",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Latest response to this webhook.",
                        max_length=8000,
                        verbose_name="last response",
                    ),
                ),
                (
                    "last_success",
                    models.DateTimeField(
                        default=None,
                        help_text="When the webhook last succeeded.",
                        null=True,
                        verbose_name="last success",
                    ),
                ),
                (
                    "last_failure",
                    models.DateTimeField(
                        default=None,
                        help_text="When the webhook last failed.",
                        null=True,
                        verbose_name="last failure",
                    ),
                ),
            ],
            options={
                "abstract": False,
                "swappable": "SIGNAL_WEBHOOKS_CUSTOM_MODEL",
            },
        ),
        migrations.AddConstraint(
            model_name="webhook",
            constraint=models.UniqueConstraint(
                fields=("ref", "endpoint"),
                name="prevent_duplicate_hooks_signal_webhooks_webhook",
            ),
        ),
    ]



# Generated by Django 4.2.6 on 2023-12-02 11:01

from django.db import migrations, models

import signal_webhooks.utils


class Migration(migrations.Migration):
    dependencies = [
        ("signal_webhooks", "0002_change_auth_token_and_last_response_max_length"),
    ]

    operations = [
        migrations.AlterField(
            model_name="webhook",
            name="endpoint",
            field=models.URLField(
                help_text="Target endpoint for this webhook.",
                max_length=2047,
                verbose_name="endpoint",
            ),
        ),
        migrations.AlterField(
            model_name="webhook",
            name="name",
            field=models.CharField(
                db_index=True,
                help_text="Webhook name.",
                max_length=255,
                unique=True,
                verbose_name="name",
            ),
        ),
        migrations.AlterField(
            model_name="webhook",
            name="ref",
            field=models.CharField(
                db_index=True,
                help_text="Dot import notation to the model the webhook is for.",
                max_length=1023,
                validators=[signal_webhooks.utils.model_from_reference],
                verbose_name="referenced model",
            ),
        ),
        migrations.AlterField(
            model_name="webhook",
            name="signal",
            field=models.CharField(
                choices=[
                    ("CREATE", "Create"),
                    ("UPDATE", "Update"),
                    ("DELETE", "Delete"),
                    ("M2M", "M2M changed"),
                    ("CREATE_OR_UPDATE", "Create or Update"),
                    ("CREATE_OR_DELETE", "Create or Delete"),
                    ("CREATE_OR_M2M", "Create or M2M changed"),
                    ("UPDATE_OR_DELETE", "Update or Delete"),
                    ("UPDATE_OR_M2M", "Update or M2M changed"),
                    ("DELETE_OR_M2M", "Delete or M2M changed"),
                    ("CREATE_UPDATE_OR_DELETE", "Create, Update or Delete"),
                    ("CREATE_UPDATE_OR_M2M", "Create, Update or M2M changed"),
                    ("CREATE_DELETE_OR_M2M", "Create, Delete or M2M changed"),
                    ("UPDATE_DELETE_OR_M2M", "Update, Delete or M2M changed"),
                    ("CREATE_UPDATE_DELETE_OR_M2M", "Create, Update or Delete, or M2M changed"),
                ],
                help_text="Signal the webhook fires to.",
                max_length=255,
                verbose_name="signal",
            ),
        ),
    ]



from __future__ import annotations

from django.apps import AppConfig

__all__ = [
    "DjangoSignalWebhooksConfig",
]


class DjangoSignalWebhooksConfig(AppConfig):
    name = "signal_webhooks"
    verbose_name = "Django Signal Webhooks"
    default_auto_field = "django.db.models.BigAutoField"



# Generated by Django 4.2.6 on 2023-11-25 08:58

from django.db import migrations, models

import signal_webhooks.fields
import signal_webhooks.utils


class Migration(migrations.Migration):
    dependencies = [
        ("signal_webhooks", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="webhook",
            name="auth_token",
            field=signal_webhooks.fields.TokenField(
                blank=True,
                default="",
                help_text="Authentication token to use in an Authorization header.",
                max_length=8000,
                validators=[signal_webhooks.utils.decode_cipher_key],
                verbose_name="authentication token",
            ),
        ),
        migrations.AlterField(
            model_name="webhook",
            name="last_response",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Latest response to this webhook.",
                max_length=8000,
                verbose_name="last response",
            ),
        ),
    ]



from __future__ import annotations

from django.test.signals import setting_changed
from settings_holder import SettingsHolder, reload_settings

from .typing import Any, HooksData, NamedTuple, Union

__all__ = [
    "webhook_settings",
]


class DefaultSettings(NamedTuple):
    #
    # Defines hooks for models. Key in the dict is the dot import path for a model,
    # e.g., "django.contrib.auth.models.User", and the value for a key defines that
    # model's hooks. Setting the value (or any of value for the keys in 'HooksData')
    # to ... (ellipses) will use the default handlers for the model (or appropriate
    # signal from 'HooksData' key). Setting the value (or any of value for the keys in
    # 'HooksData') to None will explicitly not allow hooks for that model (or appropriate
    # signal from 'HooksData' key). Webhooks cannot be created without the appropriate
    # definition in this setting.
    HOOKS: dict[str, HooksData | None] = {}
    #
    # Timeout for responses from webhooks before they fail.
    TIMEOUT: int = 10
    #
    # Cipher key to use when encrypting tokens into the database.
    # Should be 16, 24, or 32 bytes converted to base64. You can use
    # 'signal_webhooks.utils.random_cipher_key' to generate one.
    CIPHER_KEY: str | None = None
    #
    # When this is set to True, auth_token will be hidden in admin panel after
    # it has been set. A small snippet from the end of the code will be shown
    # in the helptext for the field. Saving the model again without adding another
    # token will reuse the hidden token.
    HIDE_TOKEN: bool = False
    #
    # Default serializer function to use for serializing model to json-acceptable
    # data. Takes these arguments (instance: Model), and should return
    # data matching 'signal_webhooks.typing.JSONData'.
    #
    # Can also be overridden on per-model basis by declaring a 'webhook_data'
    # method on the model. This function takes no arguments, and should return
    # data matching 'signal_webhooks.typing.JSONData'.
    SERIALIZER: str = "signal_webhooks.utils.default_serializer"
    #
    # Hook for adding additional arguments for the http client that sends the webhooks.
    # Takes these arguments (hook: Webhook), and should return data matching
    # 'signal_webhooks.typing.ClientKwargs'. Note that the headers from the hook will be
    # updated to the 'headers' argument, and the data sent by the webhook will be in json form.
    CLIENT_KWARGS: str = "signal_webhooks.utils.default_client_kwargs"
    #
    # Hook for adding additional filtering to the database query when selecting hooks to fire.
    # Takes these arguments (instance: Model, method: Literal['CREATE', 'UPDATE', 'DELETE']),
    # and should return a dict with the additional arguments passed to 'QuerySet.filter()'.
    # See 'signal_webhooks.models.WebhookQuerySet.get_for_model' for the filtering arguments
    # that are already added by default.
    FILTER_KWARGS: str = "signal_webhooks.utils.default_filter_kwargs"
    #
    # Error handing function that will be called if a webhook fails. Takes these
    # arguments (hook: Webhook, error: Optional[Exception]) and returns None.
    # "error" will be given if the webhook timed out, or a response from the
    # client could not otherwise be created. Note, that the handler will be run
    # inside an async event loop, so 'asgiref.sync_to_async' should be used for
    # any database calls.
    ERROR_HANDLER: str = "signal_webhooks.handlers.default_error_handler"
    #
    # Function that starts the hook once it has been found. Takes these arguments
    # (hook: Callable[..., None], **kwargs: Any) and returns None. The default handler
    # starts a thread that calls the hook with the given kwargs.
    TASK_HANDLER: str = "signal_webhooks.handlers.thread_task_handler"
    #
    # Unique id for the 'signals.post_save' receiver the webhooks are using.
    DISPATCH_UID_POST_SAVE: str = "django-signal-webhooks-post-save"
    #
    # Unique id for the 'signals.post_delete' receiver the webhooks are using.
    DISPATCH_UID_POST_DELETE: str = "django-signal-webhooks-post-delete"
    #
    # Unique id for the 'signals.m2m_changed' receiver the webhooks are using.
    DISPATCH_UID_M2M_CHANGED: str = "django-signal-webhooks-m2m-changed"


SETTING_NAME: str = "SIGNAL_WEBHOOKS"

DEFAULTS = DefaultSettings()._asdict()

IMPORT_STRINGS: set[Union[bytes, str]] = {
    "HOOKS",
    "SERIALIZER",
    "CLIENT_KWARGS",
    "FILTER_KWARGS",
    "ERROR_HANDLER",
    "TASK_HANDLER",
}

REMOVED_SETTINGS: set[str] = set()


class WebhookSettingsHolder(SettingsHolder):
    def make_imports(self, name: str, value: Any) -> Any:
        if name != "HOOKS":
            return super().make_imports(name, value)

        self.resolve_hooks(name, value)
        return value

    def resolve_hooks(self, name: str, value: Any) -> None:
        if not isinstance(value, dict):  # pragma: no cover
            msg = "'HOOKS' must be a dict."
            raise TypeError(msg)

        for model_path, webhooks in value.items():
            if not isinstance(model_path, str):  # pragma: no cover
                msg = "'HOOKS' keys must be strings."
                raise TypeError(msg)

            if webhooks in (..., None):
                continue

            if not isinstance(webhooks, dict):  # pragma: no cover
                msg = f"'HOOKS[{model_path}]' values must be dicts, ellipsis, or None."
                raise TypeError(msg)

            for method, func_path in webhooks.items():
                allowed_methods = ("CREATE", "UPDATE", "DELETE", "M2M_ADD", "M2M_REMOVE", "M2M_CLEAR")

                if method not in allowed_methods:  # pragma: no cover
                    msg = f"'HOOKS[{model_path}]' keys must be one of {allowed_methods}. Got {method!r}."
                    raise TypeError(msg)

                if func_path in (..., None):
                    continue

                if not isinstance(func_path, str):  # pragma: no cover
                    msg = f"'HOOKS[{model_path}]' values must be strings, ellipsis, or None. Got {func_path!r}."
                    raise TypeError(msg)

                value[model_path][method] = self.import_from_string(func_path, name)


webhook_settings = WebhookSettingsHolder(
    setting_name=SETTING_NAME,
    defaults=DEFAULTS,
    import_strings=IMPORT_STRINGS,
    removed_settings=REMOVED_SETTINGS,
)

reload_my_settings = reload_settings(SETTING_NAME, webhook_settings)
setting_changed.connect(reload_my_settings)



from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING

from django.core.serializers import python

if TYPE_CHECKING:
    from django.db import models

__all__ = [
    "webhook_serializer",
]


logger = logging.getLogger(__name__)


class _WebhookSerializer(python.Serializer):
    """Custom serializer to skip m2m fields when serializing models post-delete."""

    def handle_m2m_field(self, obj: models.Model, field: models.Field) -> None:
        try:
            super().handle_m2m_field(obj, field)
        except Exception as error:  # noqa: BLE001 pragma: no cover
            logger.debug(f"Skip {field.name!r} during post-delete signal.", exc_info=error)
            self._current[field.name] = []

    def end_serialization(self) -> None:
        # Convert any non-serializable objects to strings
        self.objects = json.loads(json.dumps(self.objects[0], default=str))


webhook_serializer = _WebhookSerializer()



from rest_framework.viewsets import ModelViewSet

from signal_webhooks.utils import get_webhook_model

from .serializers import WebhookSerializer

__all__ = [
    "WebhooksViewSet",
]


class WebhooksViewSet(ModelViewSet):
    serializer_class = WebhookSerializer
    queryset = get_webhook_model().objects.all()



from __future__ import annotations

from typing import TYPE_CHECKING

from rest_framework.exceptions import ValidationError
from rest_framework.serializers import ModelSerializer

from signal_webhooks.utils import get_webhook_model

if TYPE_CHECKING:
    from signal_webhooks.typing import Any

__all__ = [
    "WebhookSerializer",
]


class WebhookSerializer(ModelSerializer):
    class Meta:
        model = get_webhook_model()
        fields = [
            "id",
            "name",
            "signal",
            "ref",
            "endpoint",
            "headers",
            "auth_token",
            "enabled",
            "keep_last_response",
        ]

    def validate(self, attrs: dict[str, Any]) -> dict[str, Any]:
        if get_webhook_model().objects.filter(ref=attrs["ref"], endpoint=attrs["endpoint"]).exists():
            msg = "Webhook for this model to this endpoint already exists."  # pragma: no cover
            raise ValidationError(msg)  # pragma: no cover

        return attrs



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from django import forms
from django.contrib import admin

from .settings import webhook_settings
from .utils import get_webhook_model

if TYPE_CHECKING:
    from .typing import Any, Optional, Union

__all__ = [
    "WebhookAdmin",
    "WebhookModelForm",
]


logger = logging.getLogger(__name__)


WebhookModel = get_webhook_model()


class WebhookModelForm(forms.ModelForm):
    class Meta:
        model = WebhookModel
        fields = "__all__"  # noqa: DJ007
        widgets = {
            "auth_token": forms.Textarea,
        }

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if webhook_settings.HIDE_TOKEN:
            instance: Optional[WebhookModel] = kwargs.get("instance")
            self._auth_token: Optional[str] = None
            if instance is not None:
                self._auth_token = instance.auth_token
                instance.auth_token = ""

        super().__init__(*args, **kwargs)

        if webhook_settings.HIDE_TOKEN and self._auth_token is not None:
            masked_auth_token = self._auth_token.replace(self._auth_token[:-5], "********", 1)
            self.fields["auth_token"].help_text += f" Current token: {masked_auth_token}"

    def clean(self) -> Union[dict[str, Any], None]:
        if webhook_settings.HIDE_TOKEN and self.cleaned_data["auth_token"] == "" and self._auth_token is not None:
            self.cleaned_data["auth_token"] = self._auth_token
        return super().clean()


@admin.register(WebhookModel)
class WebhookAdmin(admin.ModelAdmin):
    form = WebhookModelForm
    list_display = [
        "name",
        "ref",
        "endpoint",
        "last_success",
        "last_failure",
        "enabled",
    ]
    list_filter = [
        "ref",
        "enabled",
        "signal",
    ]
    search_fields = [
        "name",
        "ref",
        "endpoint",
        "last_response",
    ]
    readonly_fields = [
        "last_response",
        "last_success",
        "last_failure",
    ]

    def lookup_allowed(self, lookup: str, value: str) -> bool:  # pragma: no cover
        # Don't allow lookups involving auth tokens
        return not lookup.startswith("auth_token") and super().lookup_allowed(lookup, value)



from __future__ import annotations

import asyncio
import datetime
import logging
from threading import Thread
from typing import TYPE_CHECKING

import httpx
from asgiref.sync import sync_to_async
from django.db import models
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver

from .exceptions import WebhookCancelled
from .settings import webhook_settings
from .typing import ACTION_TO_METHOD
from .utils import get_webhook_model, reference_for_model, tasks_as_completed, truncate

if TYPE_CHECKING:
    from django.db.models.base import ModelBase

    from .models import Webhook
    from .typing import (
        Any,
        Callable,
        ClientKwargs,
        HooksData,
        JSONData,
        M2MChangedData,
        Method,
        PostDeleteData,
        PostSaveData,
    )


__all__ = [
    "default_error_handler",
    "default_hook_handler",
    "sync_task_handler",
    "thread_task_handler",
    "webhook_delete_handler",
    "webhook_update_create_handler",
]


logger = logging.getLogger(__name__)


@receiver(post_save, dispatch_uid=webhook_settings.DISPATCH_UID_POST_SAVE)
def webhook_update_create_handler(sender: ModelBase, **kwargs: Any) -> None:  # noqa: ARG001
    kwargs: PostSaveData
    method: Method = "CREATE" if kwargs["created"] else "UPDATE"  # type: ignore[assignment]
    webhook_handler(instance=kwargs["instance"], method=method)


@receiver(post_delete, dispatch_uid=webhook_settings.DISPATCH_UID_POST_DELETE)
def webhook_delete_handler(sender: ModelBase, **kwargs: Any) -> None:  # noqa: ARG001
    kwargs: PostDeleteData
    method: Method = "DELETE"
    webhook_handler(instance=kwargs["instance"], method=method)


@receiver(models.signals.m2m_changed, dispatch_uid=webhook_settings.DISPATCH_UID_M2M_CHANGED)
def webhook_m2m_handler(sender: ModelBase, **kwargs: Any) -> None:  # noqa: ARG001
    kwargs: M2MChangedData
    method: Method | None = ACTION_TO_METHOD.get(kwargs["action"])
    # Don't fire webhooks for pre-actions
    if method is None:
        return
    webhook_handler(instance=kwargs["instance"], method=method)


def webhook_handler(instance: models.Model, method: Method) -> None:
    ref = reference_for_model(type(instance))

    hook = find_hook_handler(ref, method)
    if hook is None:
        return

    try:
        data = webhook_settings.SERIALIZER(instance)
    except WebhookCancelled as error:
        logger.info(f"{method.capitalize()} webhook for {ref!r} cancelled before it was sent. Reason given: {error}")
        return
    except Exception as error:
        logger.exception(
            f"{method.capitalize()} webhook data for {ref!r} could not be created.",
            exc_info=error,
        )
        return

    webhook_settings.TASK_HANDLER(hook, instance=instance, data=data, method=method)


def find_hook_handler(ref: str, method: Method) -> Callable | None:
    hook: Callable | None = ...
    hooks: HooksData | None = webhook_settings.HOOKS.get(ref)

    if hooks is None:
        return None
    if hooks is not ...:
        hook = hooks.get(method)
    if hook is None:
        return None
    if hook is ...:
        hook = default_hook_handler

    return hook


def default_error_handler(hook: Webhook, error: Exception | None) -> None:
    """
    Default handler for errors from webhooks.

    :param hook: Hook that failed
    :param error: Exception if one was caused due to, e.g., timeout.
    """


def thread_task_handler(hook: Callable[..., None], **kwargs: Any) -> None:
    thread = Thread(target=hook, kwargs=kwargs)
    thread.start()


def sync_task_handler(hook: Callable[..., None], **kwargs: Any) -> None:
    hook(**kwargs)


def default_hook_handler(instance: models.Model, data: JSONData, method: Method) -> None:
    hooks: models.QuerySet = get_webhook_model().objects.get_for_model(instance, method=method)
    if not hooks.exists():
        return

    client_kwargs = build_client_kwargs_by_hook_id(hooks)
    asyncio.run(fire_webhooks(hooks, data, client_kwargs))


def build_client_kwargs_by_hook_id(hooks: models.QuerySet) -> dict[int, ClientKwargs]:
    client_kwargs_by_hook_id: dict[int, ClientKwargs] = {}
    for hook in hooks:
        client_kwargs_by_hook_id[hook.id] = webhook_settings.CLIENT_KWARGS(hook)
        client_kwargs_by_hook_id[hook.id].setdefault("headers", {})
        client_kwargs_by_hook_id[hook.id]["headers"].update(hook.default_headers())
    return client_kwargs_by_hook_id


async def fire_webhooks(hooks: models.QuerySet, data: JSONData, client_kwargs: dict[int, ClientKwargs]) -> None:
    futures: set[asyncio.Task] = set()
    hooks_by_name: dict[str, Webhook] = {hook.name: hook for hook in hooks}
    webhook_model = get_webhook_model()

    async with httpx.AsyncClient(timeout=webhook_settings.TIMEOUT, follow_redirects=True) as client:
        futures.update(
            asyncio.Task(
                client.post(hook.endpoint, json=data, **client_kwargs[hook.id]),
                name=hook.name,
            )
            for hook in hooks
        )

        async for task in tasks_as_completed(futures):
            hook = hooks_by_name[task.get_name()]

            try:
                response: httpx.Response = task.result()
            except Exception as error:
                logger.exception(f"Webhook {hook.name!r} failed.", exc_info=error)
                hook.last_failure = datetime.datetime.now(tz=datetime.timezone.utc)
                webhook_settings.ERROR_HANDLER(hook, error)
                continue

            if response.status_code // 100 == 2:  # noqa: PLR2004
                hook.last_success = datetime.datetime.now(tz=datetime.timezone.utc)
                if hook.keep_last_response:
                    hook.last_response = truncate(response.content.decode())

            else:
                hook.last_failure = datetime.datetime.now(tz=datetime.timezone.utc)
                webhook_settings.ERROR_HANDLER(hook, None)
                if hook.keep_last_response:
                    hook.last_response = truncate(response.content.decode())

    await sync_to_async(webhook_model.objects.bulk_update)(
        objs=hooks,
        fields=[
            "last_success",
            "last_failure",
            "last_response",
        ],
    )



from __future__ import annotations

import base64
import logging
import os
from typing import TYPE_CHECKING

from cryptography.exceptions import InvalidTag
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from django.core.exceptions import ValidationError
from django.db import models

from .utils import decode_cipher_key

if TYPE_CHECKING:
    from .typing import Any

__all__ = [
    "TokenField",
]


logger = logging.getLogger(__name__)


class TokenField(models.CharField):
    """Encrypt token with a cipher before saving it."""

    def from_db_value(self, value: str, *args: Any, **kwargs: Any) -> str:
        if not value:
            return value

        key = decode_cipher_key()

        string = base64.b64decode(value)
        nonce = string[:12]
        data = string[12:]
        cipher = AESGCM(key=key)

        try:
            decrypted_token = cipher.decrypt(nonce, data, None)
        except InvalidTag as error:
            msg = "Wrong cipher key."
            raise ValidationError(msg) from error

        return decrypted_token.decode()

    def get_prep_value(self, value: str, *args: Any, **kwargs: Any) -> str:
        if not value:
            return value

        key = decode_cipher_key()

        nonce = os.urandom(12)
        cipher = AESGCM(key)
        encrypted_token = cipher.encrypt(nonce, value.encode(encoding="utf-8"), None)
        return base64.b64encode(nonce + encrypted_token).decode()



# ruff: noqa: ANN001, ARG001, FBT003

from django.db import migrations, models

migrations_mapping = {
    0: "CREATE",
    1: "UPDATE",
    2: "DELETE",
    3: "CREATE_OR_UPDATE",
    4: "CREATE_OR_DELETE",
    5: "UPDATE_OR_DELETE",
    6: "CREATE_UPDATE_OR_DELETE",
}


def migrate_signal_choices(apps, schema_editor) -> None:  # pragma: no cover
    # Migrate the signal choices from ints to strings.
    Webhook = apps.get_model("signal_webhooks", "Webhook")
    Webhook.objects.update(
        signal=models.Case(
            *(models.When(models.Q(signal=i), then=models.Value(signal)) for i, signal in migrations_mapping.items()),
            default=models.Value("CREATE_UPDATE_OR_DELETE"),
        ),
    )


def reverse_migrate_signal_choices(apps, schema_editor) -> None:  # pragma: no cover
    # Reverse the signal migration.
    Webhook = apps.get_model("signal_webhooks", "Webhook")
    Webhook.objects.update(
        signal=models.Case(
            *(models.When(models.Q(signal=signal), then=models.Value(i)) for i, signal in migrations_mapping.items()),
            default=models.Value(6),  # "ALL"
        ),
        # Disable all webhooks that are not in the new signal choices.
        enabled=models.Case(
            models.When(models.Q(signal__in=list(migrations_mapping.values())), then=models.Value(True)),
            default=models.Value(False),
        ),
    )


class Migration(migrations.Migration):
    dependencies = [
        ("signal_webhooks", "0003_alter_webhook_endpoint_alter_webhook_name_and_more"),
    ]

    operations = [
        migrations.RunPython(
            code=migrate_signal_choices,
            reverse_code=reverse_migrate_signal_choices,
        )
    ]



from __future__ import annotations

from typing import TYPE_CHECKING

from django.db import models

from .fields import TokenField
from .settings import webhook_settings
from .typing import MAX_COL_SIZE, METHOD_SIGNALS, SignalChoices
from .utils import decode_cipher_key, is_dict, model_from_reference, reference_for_model

if TYPE_CHECKING:
    import datetime

    from django.db.models import Model

    from .typing import Any, Method, Self


__all__ = [
    "Webhook",
    "WebhookBase",
]


class WebhookQuerySet(models.QuerySet):
    """Webhook queryset."""

    def get_for_model(self, instance: Model, method: Method) -> Self:
        kwargs: dict[str, Any] = webhook_settings.FILTER_KWARGS(instance, method)
        return self.filter(
            ref=reference_for_model(type(instance)),
            signal__in=METHOD_SIGNALS[method],
            enabled=True,
            **kwargs,
        )


class WebhookBase(models.Model):
    """Base webhook stuff."""

    name: str = models.CharField(
        unique=True,
        db_index=True,
        max_length=255,
        verbose_name="name",
        help_text="Webhook name.",
    )
    signal: SignalChoices = models.CharField(
        max_length=255,
        verbose_name="signal",
        help_text="Signal the webhook fires to.",
        choices=SignalChoices.choices,
    )
    ref: str = models.CharField(
        max_length=1023,
        db_index=True,
        verbose_name="referenced model",
        help_text="Dot import notation to the model the webhook is for.",
        validators=[model_from_reference],
    )
    endpoint: str = models.URLField(
        max_length=2047,
        verbose_name="endpoint",
        help_text="Target endpoint for this webhook.",
    )
    headers: dict[str, Any] = models.JSONField(
        blank=True,
        default=dict,
        verbose_name="headers",
        help_text="Headers to send with the webhook request.",
        validators=[is_dict],
    )
    auth_token: str = TokenField(
        default="",
        blank=True,
        max_length=MAX_COL_SIZE,
        verbose_name="authentication token",
        help_text="Authentication token to use in an Authorization header.",
        validators=[decode_cipher_key],
    )
    enabled: bool = models.BooleanField(
        default=True,
        verbose_name="enabled",
        help_text="Is this webhook enabled?",
    )
    keep_last_response: bool = models.BooleanField(
        default=False,
        verbose_name="keep last response",
        help_text="Should the webhook keep a log of the latest response it got?",
    )
    created: datetime.datetime = models.DateTimeField(
        auto_now_add=True,
        verbose_name="created",
        help_text="When the webhook was created.",
    )
    updated: datetime.datetime = models.DateTimeField(
        auto_now=True,
        verbose_name="updated",
        help_text="When the webhook was last updated.",
    )
    last_response: str = models.CharField(
        default="",
        blank=True,
        max_length=MAX_COL_SIZE,
        verbose_name="last response",
        help_text="Latest response to this webhook.",
    )
    last_success: datetime.datetime | None = models.DateTimeField(
        null=True,
        default=None,
        verbose_name="last success",
        help_text="When the webhook last succeeded.",
    )
    last_failure: datetime.datetime | None = models.DateTimeField(
        null=True,
        default=None,
        verbose_name="last failure",
        help_text="When the webhook last failed.",
    )

    objects = WebhookQuerySet.as_manager()

    class Meta:
        abstract = True
        constraints = [
            models.UniqueConstraint(
                fields=["ref", "endpoint"],
                name="prevent_duplicate_hooks_%(app_label)s_%(class)s",
            ),
        ]

    def __str__(self) -> str:
        return self.name

    def default_headers(self) -> dict[str, str]:
        headers = self.headers.copy()
        headers.setdefault("Content-Type", "application/json")
        if self.auth_token:
            headers.setdefault("Authorization", self.auth_token)

        return headers


class Webhook(WebhookBase):
    """Saved webhooks."""

    class Meta(WebhookBase.Meta):
        swappable = "SIGNAL_WEBHOOKS_CUSTOM_MODEL"



from __future__ import annotations

import asyncio
import base64
import logging
import os
import sys
from functools import cache
from importlib import import_module
from typing import TYPE_CHECKING

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db.models.base import ModelBase

from .serializers import webhook_serializer
from .settings import webhook_settings
from .typing import MAX_COL_SIZE, ClientKwargs

if TYPE_CHECKING:
    from django.db.models import Model

    from .models import Webhook, WebhookBase
    from .typing import Any, Generator, JSONData, Literal, Method


__all__ = [
    "decode_cipher_key",
    "default_client_kwargs",
    "default_serializer",
    "get_webhook_model",
    "is_dict",
    "model_from_reference",
    "random_cipher_key",
    "reference_for_model",
    "tasks_as_completed",
    "truncate",
]


logger = logging.getLogger(__name__)


def is_dict(value: str) -> None:
    if not isinstance(value, dict):
        msg = "Headers should always be a dict."
        raise ValidationError(msg)


def truncate(string: str, limit: int = MAX_COL_SIZE) -> str:
    if len(string) > limit:
        string = string[: limit - 3] + "..."
    return string


def random_cipher_key(length: Literal[16, 24, 32] = 16) -> str:
    return base64.b64encode(os.urandom(length)).decode()


# 'value' as parameter so that can be used as a field validator
def decode_cipher_key(value: str = "") -> bytes:  # noqa: ARG001
    try:
        return base64.b64decode(webhook_settings.CIPHER_KEY)
    except TypeError as error:
        msg = "Cipher key not set."
        raise ValidationError(msg) from error
    except Exception as error:
        msg = "Invalid cipher key."
        raise ValidationError(msg) from error


def default_serializer(instance: Model) -> JSONData:
    if hasattr(instance, "webhook_data") and callable(instance.webhook_data):
        return instance.webhook_data()

    return webhook_serializer.serialize([instance])


def default_client_kwargs(hook: Webhook) -> ClientKwargs:  # noqa: ARG001
    return ClientKwargs()


def default_filter_kwargs(instance: Model, method: Method) -> dict[str, Any]:  # noqa: ARG001
    return {}


@cache
def get_webhook_model() -> type[WebhookBase]:
    ref = getattr(settings, "SIGNAL_WEBHOOKS_CUSTOM_MODEL", "signal_webhooks.models.Webhook")

    if ref == "signal_webhooks.models.Webhook":
        from .models import Webhook

        return Webhook

    try:
        model = model_from_reference(ref, check_hooks=False)
    except ValidationError as error:
        msg = f"{ref!r} is not a model that can be imported."
        raise ImproperlyConfigured(msg) from error

    from .models import WebhookBase

    if not issubclass(model, WebhookBase):
        base_ref = reference_for_model(WebhookBase)
        msg = f"{ref!r} is not a subclass of a {base_ref!r}."
        raise ImproperlyConfigured(msg)

    return model


def model_from_reference(ref: str, check_hooks: bool = True) -> ModelBase:  # noqa: FBT001, FBT002
    msg = f"Could not import {ref!r}"
    try:
        module_path, class_name = ref.rsplit(".", 1)
    except ValueError as error:
        new_msg = f"{msg}. {ref!r} doesn't look like a module path."
        raise ValidationError(new_msg) from error

    if module_path not in sys.modules or (
        # Module is not fully initialized.
        getattr(sys.modules[module_path], "__spec__", None) is not None
        and getattr(sys.modules[module_path].__spec__, "_initializing", False) is True
    ):  # pragma: no cover
        try:
            import_module(module_path)
        except ImportError as error:
            new_msg = f"{msg}. {error.__class__.__name__!r}: {error}."
            raise ValidationError(new_msg) from error

    try:
        model_type = getattr(sys.modules[module_path], class_name)
    except AttributeError as error:
        new_msg = f"{msg}. Module {module_path!r} does not define {class_name!r}."
        raise ValidationError(new_msg) from error

    if not isinstance(model_type, ModelBase):
        new_msg = f"{ref!r} is not a django model."
        raise ValidationError(new_msg)

    if check_hooks and ref not in webhook_settings.HOOKS:
        new_msg = f"Webhooks not defined for {ref!r}."
        raise ValidationError(new_msg)

    return model_type


def reference_for_model(model: ModelBase) -> str:
    return f"{model.__module__}.{model.__name__}"


async def tasks_as_completed(tasks: set[asyncio.Task]) -> Generator[asyncio.Task, Any, None]:
    done = asyncio.Queue()

    def _on_completion(task_: asyncio.Task) -> None:
        tasks.remove(task_)
        done.put_nowait(task_)

    async def _wait_for_one() -> asyncio.Task:
        return await done.get()

    for task in tasks:
        task.add_done_callback(_on_completion)

    for _ in range(len(tasks)):
        yield await _wait_for_one()



from __future__ import annotations

from collections.abc import Callable, Coroutine, Generator, Iterator, Mapping, Sequence
from typing import TYPE_CHECKING, Any, Literal, NamedTuple, TypedDict, Union

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

import django
from django.db import models

if TYPE_CHECKING:
    from django.db.models import QuerySet
    from django.db.models.base import Model
    from django.db.models.signals import ModelSignal
    from httpx._client import UseClientDefault
    from httpx._types import (
        AuthTypes,
        CookieTypes,
        HeaderTypes,
        QueryParamTypes,
        RequestContent,
        RequestData,
        RequestFiles,
        TimeoutTypes,
    )

__all__ = [
    "Any",
    "Callable",
    "ClientKwargs",
    "Coroutine",
    "Generator",
    "HooksData",
    "Iterator",
    "JSONData",
    "JSONValue",
    "Literal",
    "Method",
    "NamedTuple",
    "PostDeleteData",
    "PostSaveData",
    "Self",
    "Sequence",
    "SignalChoices",
    "TypedDict",
    "Union",
]

JSONValue = Union[str, int, float, bool, None, tuple["JSONValue"], list["JSONValue"], dict[str, "JSONValue"]]
JSONData = Union[list[dict[str, JSONValue]], dict[str, JSONValue]]
Method = Literal["CREATE", "UPDATE", "DELETE", "M2M_ADD", "M2M_REMOVE", "M2M_CLEAR"]
M2MAction = Literal["pre_add", "post_add", "pre_remove", "post_remove", "pre_clear", "post_clear"]


class PostSaveData(TypedDict):
    signal: ModelSignal
    instance: Model
    created: bool
    raw: bool
    using: str  # database name
    update_fields: Union[frozenset[str], None]


class PostDeleteData(TypedDict):
    signal: ModelSignal
    instance: Model
    using: str  # database name

    if django.VERSION >= (4, 1):
        origin: Union[Model, QuerySet]


class M2MChangedData(TypedDict):
    signal: ModelSignal
    action: M2MAction
    instance: Model
    reverse: bool
    model: type[Model]
    pk_set: set[int]
    using: str  # database name


class ClientKwargs(TypedDict, total=False):
    content: RequestContent
    data: RequestData
    files: RequestFiles
    json: JSONData
    params: QueryParamTypes
    headers: HeaderTypes
    cookies: CookieTypes
    auth: Union[AuthTypes, UseClientDefault]
    follow_redirects: Union[bool, UseClientDefault]
    timeout: Union[TimeoutTypes, UseClientDefault]
    extensions: Mapping[str, Any]


class HooksData(TypedDict, total=False):
    CREATE: Union[str, Callable, None]
    UPDATE: Union[str, Callable, None]
    DELETE: Union[str, Callable, None]
    M2M_ADD: Union[str, Callable, None]
    M2M_REMOVE: Union[str, Callable, None]
    M2M_CLEAR: Union[str, Callable, None]


class SignalChoices(models.TextChoices):
    CREATE = (
        "CREATE",
        "Create",
    )
    UPDATE = (
        "UPDATE",
        "Update",
    )
    DELETE = (
        "DELETE",
        "Delete",
    )
    M2M = (
        "M2M",
        "M2M changed",
    )
    CREATE_OR_UPDATE = (
        "CREATE_OR_UPDATE",
        "Create or Update",
    )
    CREATE_OR_DELETE = (
        "CREATE_OR_DELETE",
        "Create or Delete",
    )
    CREATE_OR_M2M = (
        "CREATE_OR_M2M",
        "Create or M2M changed",
    )
    UPDATE_OR_DELETE = (
        "UPDATE_OR_DELETE",
        "Update or Delete",
    )
    UPDATE_OR_M2M = (
        "UPDATE_OR_M2M",
        "Update or M2M changed",
    )
    DELETE_OR_M2M = (
        "DELETE_OR_M2M",
        "Delete or M2M changed",
    )
    CREATE_UPDATE_OR_DELETE = (
        "CREATE_UPDATE_OR_DELETE",
        "Create, Update or Delete",
    )
    CREATE_UPDATE_OR_M2M = (
        "CREATE_UPDATE_OR_M2M",
        "Create, Update or M2M changed",
    )
    CREATE_DELETE_OR_M2M = (
        "CREATE_DELETE_OR_M2M",
        "Create, Delete or M2M changed",
    )
    UPDATE_DELETE_OR_M2M = (
        "UPDATE_DELETE_OR_M2M",
        "Update, Delete or M2M changed",
    )
    CREATE_UPDATE_DELETE_OR_M2M = (
        "CREATE_UPDATE_DELETE_OR_M2M",
        "Create, Update or Delete, or M2M changed",
    )

    @classmethod
    def create_choices(cls) -> set[SignalChoices]:
        return {
            cls.CREATE,
            cls.CREATE_OR_UPDATE,
            cls.CREATE_OR_DELETE,
            cls.CREATE_OR_M2M,
            cls.CREATE_UPDATE_OR_DELETE,
            cls.CREATE_DELETE_OR_M2M,
            cls.CREATE_UPDATE_OR_M2M,
            cls.CREATE_UPDATE_DELETE_OR_M2M,
        }

    @classmethod
    def update_choices(cls) -> set[SignalChoices]:
        return {
            cls.UPDATE,
            cls.CREATE_OR_UPDATE,
            cls.UPDATE_OR_DELETE,
            cls.UPDATE_OR_M2M,
            cls.CREATE_UPDATE_OR_DELETE,
            cls.CREATE_UPDATE_OR_M2M,
            cls.UPDATE_DELETE_OR_M2M,
            cls.CREATE_UPDATE_DELETE_OR_M2M,
        }

    @classmethod
    def delete_choices(cls) -> set[SignalChoices]:
        return {
            cls.DELETE,
            cls.CREATE_OR_DELETE,
            cls.UPDATE_OR_DELETE,
            cls.DELETE_OR_M2M,
            cls.CREATE_UPDATE_OR_DELETE,
            cls.CREATE_DELETE_OR_M2M,
            cls.UPDATE_DELETE_OR_M2M,
            cls.CREATE_UPDATE_DELETE_OR_M2M,
        }

    @classmethod
    def m2m_choices(cls) -> set[SignalChoices]:
        return {
            cls.M2M,
            cls.CREATE_OR_M2M,
            cls.UPDATE_OR_M2M,
            cls.DELETE_OR_M2M,
            cls.CREATE_DELETE_OR_M2M,
            cls.CREATE_UPDATE_OR_M2M,
            cls.UPDATE_DELETE_OR_M2M,
            cls.CREATE_UPDATE_DELETE_OR_M2M,
        }


METHOD_SIGNALS: dict[Method, set[SignalChoices]] = {
    "CREATE": SignalChoices.create_choices(),
    "UPDATE": SignalChoices.update_choices(),
    "DELETE": SignalChoices.delete_choices(),
    "M2M_ADD": SignalChoices.m2m_choices(),
    "M2M_REMOVE": SignalChoices.m2m_choices(),
    "M2M_CLEAR": SignalChoices.m2m_choices(),
}

ACTION_TO_METHOD: dict[M2MAction, Method] = {
    "post_add": "M2M_ADD",
    "post_remove": "M2M_REMOVE",
    "post_clear": "M2M_CLEAR",
}

# Define a maximum column size based on MS SQL Server limitation:
# https://docs.microsoft.com/en-us/sql/sql-server/maximum-capacity-specifications-for-sql-server
MAX_COL_SIZE: int = 8_000
